// Author: Eric Crahen <crahen@amazon.com>
//
// $LastChangedBy: $
// $LastChangedRevision: $
// $Date: $
// $HeadURL: $

package com.amazon.coral.value.json;

import java.io.IOException;
import java.io.StringReader;

%%

%type TokenType
%class JsonScanner
%int
%switch
%unicode
%apiprivate

%{

  static class ScannerException extends IOException {
    ScannerException(String message) { super(message); }
  }

  public static enum TokenType {
    EOF,
    NULL,
    TRUE,
    FALSE,
    NUMBER,
    STRING,
    COMMA,
    COLON,
    LEFT_BRACKET,
    RIGHT_BRACKET,
    LEFT_BRACE,
    RIGHT_BRACE
  }

  private final StringBuilder string = new StringBuilder();

  private final TokenType context[] = new TokenType[6];
  public String token, previousToken = null;
  public TokenType last, previousLast = null;

  private void fail(String msg) throws IOException {
    String ctx = "";
    for(TokenType token : context) {
      if(token == null)
        continue;
      if(ctx.length() != 0)
        ctx += " ";
      ctx += token.toString(); 
    }
    throw new ScannerException(msg + " context: " + ctx);
  }

  JsonScanner(CharSequence s) {
    this(new StringReader(s.toString()));
  }

  public void push() {
    previousLast = last;
    previousToken = token;
  }

  public TokenType next() throws IOException {
    if(previousLast != null) {
      last = previousLast;
      token = previousToken;
      previousLast = null;
      previousToken = null;
    } else {
      last = yylex();
    }
    return last;
  }

  private void context(TokenType token) {
    for(int i = 1; i < context.length; i++)
      context[i-1] = context[i];
    context[context.length-1] = token;
  }

%}

%eofval{
if(yystate() == STRING)
  fail("Unterminated string");
return TokenType.EOF;
%eofval}

%eof{ 
%eof}

%eofthrow{
IOException
%eofthrow}
/* Whitespace values */
Whitespace = [\u0009\u000A\u000D\u0020]*
MustEscape = [\u0000\u0001\u0002\u0003\u0004\u0005\u0006\u0007\u0008\u0009\n\u000B\u000C\r\u000E\u000F\u0010\u0011\u0012\u0013\u0014\u0015\u0016\u0017\u0018\u0019\u001A\u001B\u001C\u001D\u001E\u001F]

/* sval values */
_Character = [^\"\\\u0000\u0001\u0002\u0003\u0004\u0005\u0006\u0007\u0008\u0009\n\u000B\u000C\r\u000E\u000F\u0010\u0011\u0012\u0013\u0014\u0015\u0016\u0017\u0018\u0019\u001A\u001B\u001C\u001D\u001E\u001F]
Character  = {_Character}+
_HexDigit  = [0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F]

/* Number values */
Number    = [-]? {_Integral} {_Fraction}? {_Exponent}?

_Integral = ( 0 | ([1-9][0-9]*))
_Fraction = \u002E [0-9]+
_Exponent = [eE] [+-]? [0-9]+

%state YYINITIAL, STRING

%%

<YYINITIAL> {

  {Whitespace}      { }

  "true" { 
    token = "true";
    context(TokenType.TRUE);
    return TokenType.TRUE;
  }

  "false" { 
    token = "false";
    context(TokenType.FALSE);
    return TokenType.FALSE; 
  }
  
  "null" { 
    token = "null";
    context(TokenType.NULL);
    return TokenType.NULL; 
  }
  
  {Number} {
    token=yytext();
    context(TokenType.NUMBER);
    return TokenType.NUMBER;
  }

  [,] {
    token=yytext();
    context(TokenType.COMMA);
    return TokenType.COMMA;
  }

  [:] {
    token=yytext();
    context(TokenType.COLON);
    return TokenType.COLON;
  }

  [\[] {
    token=yytext();
    context(TokenType.LEFT_BRACKET);
    return TokenType.LEFT_BRACKET;
  }

  [\]] {
    token=yytext();
    context(TokenType.RIGHT_BRACKET);
    return TokenType.RIGHT_BRACKET;
  }

  [{] {
    token=yytext();
    context(TokenType.RIGHT_BRACKET);
    return TokenType.LEFT_BRACE;
  }

  [}] {
    token=yytext();
    context(TokenType.RIGHT_BRACE);
    return TokenType.RIGHT_BRACE;
  }

  [\"] { 
    yybegin(STRING);
  }

  . { fail("Unexpected character '" + yytext() + "'"); }

}
<STRING> {

  [\"] { 
    yybegin(YYINITIAL);
    token = string.toString(); 
    string.setLength(0); 
    context(TokenType.STRING);
    return TokenType.STRING; 
  }

  // Section 2.5 of RFC4627, lower control characters
  // must be escaped
  {MustEscape}+ { fail("String contains characters which must be escaped"); }

  {Character}+ { 
    /* TODO: Should these unicode verification semantics be in here?
    String s = yytext();
    for (int i = 0; i < s.length(); ++i) {
      int val = yytext().codePointAt(i);
      if (!Character.isDefined(val))
        fail(Integer.toHexString(val) + " is not a unicode codepoint.");
      //TODO validate high-low surrogates
    }
    */
    string.append( yytext() );
  }

  \\b  { string.append( '\b' ); }
  \\t  { string.append( '\t' ); }
  \\n  { string.append( '\n' ); }
  \\f  { string.append( '\f' ); }
  \\r  { string.append( '\r' ); }
  \\\" { string.append( '\"' ); }
  \\\/  { string.append( '/' ); }
  \\\\ { string.append( '\\' ); }
  \\   { fail("Incomplete escape sequence"); }
  \\u  {_HexDigit} { 
    int val = Integer.parseInt(yytext().substring(2), 16);
    // Codepoints above U+FFFF can't happen since the max codepoint you can
    // hex-escape in JSON is U+FFFF.  Other unusual codepoints that could 
    // happen here are those reserved for UTF16 surrogate pairs 
    // (U+D800 to U+DFFF).  We could reject these, but don't currently for
    // backwards compatibility.  

    string.append( (char)val );
  }

  \\. { fail("Invalid escape sequence in '" + yytext() + "'"); }

}
